@@ -1,7 +1,7 @@
 /* Simple NTP client
  *
  * Copyright (C) 1997-2015  Larry Doolittle <larry@doolittle.boa.org>
- * Copyright (C) 2010-2018  Joachim Nilsson <troglobit@gmail.com>
+ * Copyright (C) 2010-2022  Joachim Wiberg <troglobit@gmail.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License (Version 2,
@@ -27,6 +27,7 @@
 
 #include "config.h"
 #include <getopt.h>
+#include <resolv.h>
 #include <signal.h>
 #include <netdb.h>		/* getaddrinfo -> gethostbyname */
 #include <time.h>
@@ -42,8 +43,8 @@ int daemonize = 0;
 int logging = 1;
 
 const char *prognm = PACKAGE_NAME;
-static int sighup  = 0;
-static int sigterm = 0;
+static volatile sig_atomic_t sighup  = 0;
+static volatile sig_atomic_t sigterm = 0;
 
 struct ntp_peers peer;
 double root_delay;
@@ -162,7 +163,7 @@ void get_packet_timestamp(int usd, struct ntptime *udp_arrival_ntp)
 #endif
 }
 
-static int check_source(int data_len, struct sockaddr_storage *sa_source, struct ntp_control *ntpc)
+static int check_source(int data_len, struct sockaddr_storage *ss, struct ntp_control *ntpc)
 {
 	struct sockaddr_in6 *ipv6;
 	struct sockaddr_in *ipv4;
@@ -174,14 +175,14 @@ static int check_source(int data_len, struct sockaddr_storage *sa_source, struct
 	DBG("packet of length %d received", data_len);
 #endif
 
-	if (sa_source->ss_family == AF_INET) {
-		ipv4 = (struct sockaddr_in *)(sa_source);
+	if (ss->ss_family == AF_INET) {
+		ipv4 = (struct sockaddr_in *)ss;
 		port = ntohs(ipv4->sin_port);
-	} else if (sa_source->ss_family == AF_INET6) {
-		ipv6 = (struct sockaddr_in6 *)(sa_source);
+	} else if (ss->ss_family == AF_INET6) {
+		ipv6 = (struct sockaddr_in6 *)ss;
 		port = ntohs(ipv6->sin6_port);
 	} else {
-		DBG("%s: Unsupported address family", __func__);
+		ERR(0, "%s: Unsupported address family %d", __func__, ss->ss_family);
 		return 1;
 	}
 
@@ -391,7 +392,7 @@ static int rfc1305print(uint32_t *data, struct ntptime *arrival, struct ntp_cont
 	return 1;
 }
 
-void setup_receive(int usd, sa_family_t sin_family, uint16_t port)
+int setup_receive(int usd, sa_family_t sin_family, uint16_t port)
 {
 	struct sockaddr_in6 sin6;
 	struct sockaddr_in sin;
@@ -420,56 +421,57 @@ void setup_receive(int usd, sa_family_t sin_family, uint16_t port)
 
 	if (bind(usd, sa, len) == -1) {
 		ERR(errno, "Failed binding to UDP port %u", port);
-		exit(1);
+		return -1;
 	}
-	/* listen(usd,3); this isn't TCP; thanks Alexander! */
+
+	return 0;
 }
 
-static void setup_transmit(int usd, struct sockaddr_storage *ssp, uint16_t port, struct ntp_control *ntpc)
+static int setup_transmit(int usd, struct sockaddr_storage *ss, uint16_t port,
+			  struct ntp_control *ntpc)
 {
 	struct sockaddr_in6 *ipv6;
 	struct sockaddr_in *ipv4;
 	socklen_t len = 0;
 
 	/* Prefer IPv4 over IPv6, for now */
-	if (ssp->ss_family == AF_INET) {
-		ipv4 = (struct sockaddr_in *)ssp;
+	if (ss->ss_family == AF_INET) {
+		ipv4 = (struct sockaddr_in *)ss;
 		ipv4->sin_port = htons(port);
 		len = sizeof(struct sockaddr_in);
-	} else if (ssp->ss_family == AF_INET6) {
-		ipv6 = (struct sockaddr_in6 *)ssp;
+	} else if (ss->ss_family == AF_INET6) {
+		ipv6 = (struct sockaddr_in6 *)ss;
 		ipv6->sin6_port = htons(port);
 		len = sizeof(struct sockaddr_in6);
 	} else {
-		ERR(0, "%s: Unsupported address family", __func__);
-		exit(1);
+		ERR(0, "%s: Unsupported address family %d", __func__, ss->ss_family);
+		return -1;
 	}
 
-	while (connect(usd, (struct sockaddr *)ssp, len) == -1) {
-		if (ntpc->live) {
-			/* Wait here a while, networking is probably not up yet. */
-			sleep(1);
-			continue;
-		}
-
+	if (connect(usd, (struct sockaddr *)ss, len) == -1) {
 		ERR(errno, "Failed connecting to NTP server");
-		exit(1);
+		return -1;
 	}
+
+	INFO("Connected to NTP server.");
+	return 0;
 }
 
 static int getaddrbyname(char *host, struct sockaddr_storage *ss)
 {
-	int err;
+	struct addrinfo *result;
 	static int netdown = 0;
 	struct addrinfo hints;
-	struct addrinfo *result;
 	struct addrinfo *rp;
+	int err;
 
 	if (!host || !ss) {
 		errno = EINVAL;
 		return 1;
 	}
 
+	res_init();
+
 	memset(&hints, 0, sizeof(struct addrinfo));
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = 0;
@@ -482,7 +484,17 @@ static int getaddrbyname(char *host, struct sockaddr_storage *ss)
 	memset(ss, 0, sizeof(struct sockaddr_storage));
 	err = getaddrinfo(host, NULL, &hints, &result);
 	if (err) {
-		ERR(0, "Failed resolving address to hostname %s: %s", host, gai_strerror(err));
+		switch (err) {
+		case EAI_NONAME:
+			LOG("Failed resolving %s, will try again later ...", host);
+			break;
+		case EAI_AGAIN:
+			LOG("Temporary failure resolving %s, will try again later ...", host);
+			break;
+		default:
+			LOG("Error %d resolving %s: %s", err, host, gai_strerror(err));
+			break;
+		}
 		netdown = errno = ENETDOWN;
 		return 1;
 	}
@@ -521,31 +533,30 @@ static int setup_socket(struct ntp_control *ntpc)
 	struct sockaddr_storage ss;
 	int sd;
 
-	while (getaddrbyname(ntpc->server, &ss)) {
-		if (EINVAL != errno && ntpc->live) {
-			/* Wait here a while, networking is probably not up yet. */
-			sleep(1);
-			continue;
-		}
+	if (getaddrbyname(ntpc->server, &ss)) {
+		if (EINVAL != errno)
+			return -1;
 
-		ERR(0, "Unable lookup %s", ntpc->server);
+		ERR(0, "Unable to look up %s address", ntpc->server);
 		exit(1);
 	}
 
 	/* open socket based on the server address family */
-	if (ss.ss_family == AF_INET || ss.ss_family == AF_INET6) {
-		sd = socket(ss.ss_family, SOCK_DGRAM, IPPROTO_UDP);
-	} else {
-		ERR(0, "%s: Unsupported address family", __func__);
+	if (ss.ss_family != AF_INET && ss.ss_family != AF_INET6) {
+		ERR(0, "%s: Unsupported address family %d", __func__, ss.ss_family);
 		exit(1);
 	}
 
-	if (sd == -1) {
-	    return -1;
-	}
+	sd = socket(ss.ss_family, SOCK_DGRAM, IPPROTO_UDP);
+	if (sd == -1)
+		return -1;
 
-	setup_receive(sd, ss.ss_family, ntpc->local_udp_port);
-	setup_transmit(sd, &ss, ntpc->udp_port, ntpc);
+	if (setup_receive(sd, ss.ss_family, ntpc->local_udp_port) ||
+	    setup_transmit(sd, &ss, ntpc->udp_port, ntpc)) {
+		close(sd);
+		errno = ENETDOWN;
+		return -1;
+	}
 
 	/*
 	 * Every day: reopen socket and perform a new DNS lookup.
@@ -640,6 +651,11 @@ static void loop(struct ntp_control *ntpc)
 
 			usd = setup_socket(ntpc);
 			if (usd == -1) {
+				/* Wait here a while, networking is probably not up yet. */
+				if (errno == ENETDOWN) {
+					sleep(1);
+					continue;
+				}
 				ERR(errno, init ? "Failed creating UDP socket()"
 				    : "Failed reopening NTP socket");
 				goto done;
